@@ -2,113 +2,57 @@
 # @upsetjs/r
 # https://github.com/upsetjs/upsetjs_r
 #
-# Copyright (c) 2020 Samuel Gratzl <sam@sgratzl.com>
+# Copyright (c) 2021 Samuel Gratzl <sam@sgratzl.com>
 #
 
-
-sortSets = function(sets,
-                    order.by = 'cardinality',
-                    limit = NULL) {
-  set_attr = function(order.by.attr) {
-    if (order.by.attr == 'cardinality') {
-      sapply(sets, function (x)
-        if (length(x$elems) == 0)
-          x$cardinality * -1
-        else
-          length(x$elems) * -1
-        )
-    } else if (order.by.attr == 'degree') {
-      sapply(sets, function (x) length(x$setNames))
-    } else {
-      sapply(sets, function (x) x$name)
-    }
-  }
-
-  if (length(order.by) == 1 && order.by[1] == 'cardinality') {
-    order.by = c('cardinality', 'name')
-  } else if (length(order.by) == 1 && order.by[1] == 'degree') {
-    order.by = c('degree', 'name')
-  }
-  if (length(sets) > 1) {
-    values = lapply(order.by, set_attr)
-    o = do.call(order, values)
-    r = sets[o]
-  } else {
-    r = sets
-  }
-  if (is.null(limit) || length(r) <= limit) {
-    r
-  } else {
-    r[1:limit]
-  }
+#'
+#' creates a new UpSet set structure
+#' @param name name of the set
+#' @param elems the elements of the set
+#' @param cardinality the cardinality of the set, default to `length(elems)`
+#' @param color the color of the set
+#' @return the set object
+#' @examples
+#' asSet('a', c(1,2,3))
+#'
+#' @export
+asSet = function(name, elems = c(), cardinality = length(elems), color = NULL) {
+  structure(list(
+      name = name,
+      type = 'set',
+      elems = elems,
+      cardinality = cardinality,
+      color = color
+    ),
+    class = "upsetjs_set")
 }
 
-colorLookup = function(colors = NULL) {
-  if (is.null(colors)) {
-    function (c) { NULL }
-  } else {
-    color_names = names(colors)
-    function (c) { if (c %in% color_names) { colors[[c]] } else { NULL } }
-  }
-}
-
-generateCombinationsImpl = function(sets,
-                                    c_type,
-                                    min,
-                                    max,
-                                    empty,
-                                    order.by,
-                                    limit,
-                                    colors = NULL,
-                                    symbol = "&") {
-  combinations = list()
-  set_f = if (c_type == "union")
-    union
-  else
-    intersect
-  distinct = (c_type == 'distinctIntersection')
-  lsets = length(sets)
-  all_indices = 1:lsets
-  cc = colorLookup(colors)
-
-  for (l in min:(if (is.null(max)) lsets else max)) {
-    combos = combn(all_indices, l, simplify = FALSE)
-    for (combo in combos) {
-      indices = unlist(combo)
-      set_names = sapply(indices, function(i) sets[[i]]$name)
-      if (is.list(set_names)) {
-        set_names = unlist(set_names)
-      }
-      if (length(indices) == 0) {
-        elems = c()
-      } else {
-        elems = sets[[indices[1]]]$elems
-        for (index in indices) {
-          elems = set_f(elems, sets[[index]]$elems)
-        }
-      }
-      if (distinct) {
-        not_indices = setdiff(all_indices, indices)
-        for (index in not_indices) {
-          elems = setdiff(elems, sets[[index]]$elems)
-        }
-      }
-      if (empty || length(elems) > 0) {
-        c_name = paste(set_names, collapse = symbol)
-        combination = structure(list(
-          name = c_name,
-          color = cc(c_name),
-          type = c_type,
-          elems = elems,
-          setNames = set_names
-        ),
-        class = "upsetjs_combination")
-        combinations = c(combinations, list(combination))
-      }
-    }
-  }
-  names(combinations) = NULL
-  sortSets(combinations, order.by, limit)
+#'
+#' creates a new UpSet set combination structure
+#' @param name name of the set combination
+#' @param elems the elements of the set combination
+#' @param type the set combination type (intersection,distinctIntersection,union,combination)
+#' @param sets the sets this combination is part of
+#' @param cardinality the cardinality of the set, default to `length(elems)`
+#' @param color the color of the set
+#' @return the set object
+#' @examples
+#' asCombination('a', c(1,2,3))
+#'
+#' @export
+asCombination = function(name, elems = c(), type = 'intersection', sets = strsplit(name, '&'), cardinality = length(elems), color = NULL) {
+  structure(
+      list(
+        name = name,
+        type = type,
+        elems = elems,
+        color = color,
+        cardinality = cardinality,
+        setNames = sets,
+        degree = length(sets)
+      ),
+      class = "upsetjs_combination"
+    )
 }
 
 #'
@@ -120,6 +64,7 @@ generateCombinationsImpl = function(sets,
 #' @param shared a crosstalk shared data frame
 #' @param shared.mode whether on 'hover' or 'click' (default) is synced
 #' @param colors the optional list with set name to color
+#' @param c_type the combination type to use or "none" for disabling initial generation
 #' @return the object given as first argument
 #' @examples
 #' upsetjs() %>% fromList(list(a=c(1,2,3), b=c(2,3)))
@@ -131,7 +76,8 @@ fromList = function(upsetjs,
                     limit = NULL,
                     shared = NULL,
                     shared.mode = "click",
-                    colors = NULL) {
+                    colors = NULL,
+                    c_type = NULL) {
   checkUpSetCommonArgument(upsetjs)
   stopifnot(is.list(value))
   stopifnot(order.by == "cardinality" || order.by == "degree")
@@ -139,25 +85,23 @@ fromList = function(upsetjs,
               (is.numeric(limit) && length(limit) == 1))
   stopifnot(shared.mode == "click" || shared.mode == "hover")
   stopifnot(is.null(colors) || is.list(colors))
+  stopifnot(
+    is.null(c_type) ||
+      c_type == "intersection" ||
+      c_type == "union" || c_type == "distinctIntersection" ||
+      c_type == "none"
+  )
 
   elems = c()
   cc = colorLookup(colors)
   toSet = function(key, value) {
     elems <<- unique(c(elems, value))
-    structure(list(
-      name = key,
-      type = 'set',
-      elems = value,
-      color = cc(key)
-    ),
-    class = "upsetjs_set")
+    asSet(key, value, color=cc(key))
   }
-  sets = mapply(
-    toSet,
-    key = names(value),
-    value = value,
-    SIMPLIFY = FALSE
-  )
+  sets = mapply(toSet,
+                key = names(value),
+                value = value,
+                SIMPLIFY = FALSE)
   # list of list objects
   names(sets) = NULL
   names(elems) = NULL
@@ -167,21 +111,42 @@ fromList = function(upsetjs,
   }
 
   sorted_sets = sortSets(sets, order.by = order.by, limit = limit)
-  gen_sets = if (is.null(limit)) sets else sorted_sets
 
-  gen = if (isVennDiagram(upsetjs) || isKarnaughMap(upsetjs)) {
-    generateCombinationsImpl(gen_sets, 'distinctIntersection', 0, NULL, TRUE, 'degree', NULL, colors)
+  gen = if (!is.null(c_type) && c_type == "none") {
+    list() 
+  } else if (isVennDiagram(upsetjs) || isKarnaughMap(upsetjs)) {
+    generateCombinationsImpl(
+      sorted_sets,
+      ifelse(is.null(c_type), 'distinctIntersection', c_type),
+      0,
+      NULL,
+      TRUE,
+      'degree',
+      limit,
+      colors
+    )
   } else {
-    generateCombinationsImpl(gen_sets, 'intersection', 0, NULL, FALSE, order.by, NULL, colors)
+    generateCombinationsImpl(
+      sorted_sets,
+      ifelse(is.null(c_type), 'intersection', c_type),
+      0,
+      NULL,
+      FALSE,
+      order.by,
+      limit,
+      colors
+    )
   }
-  setProperties(upsetjs,
-                list(
-                  sets = sorted_sets,
-                  combinations = gen,
-                  elems = elems,
-                  expressionData = FALSE,
-                  attrs = list()
-                ))
+  setProperties(
+    upsetjs,
+    list(
+      sets = sorted_sets,
+      combinations = gen,
+      elems = elems,
+      expressionData = FALSE,
+      attrs = list()
+    )
+  )
 }
 
 #'
@@ -207,7 +172,8 @@ fromExpression = function(upsetjs,
   stopifnot(is.list(value))
   stopifnot(order.by == "cardinality" || order.by == "degree")
   stopifnot(is.null(colors) || is.list(colors))
-  stopifnot(type == "intersection" || type == "union" || type == "distinctIntersection")
+  stopifnot(type == "intersection" ||
+              type == "union" || type == "distinctIntersection")
 
   cc = colorLookup(colors)
 
@@ -218,17 +184,7 @@ fromExpression = function(upsetjs,
   raw_combinations = value
 
   toCombination = function(key, value, color) {
-    structure(
-      list(
-        name = key,
-        type = type,
-        elems = c(),
-        color = cc(key),
-        cardinality = value,
-        setNames = unlist(strsplit(key, symbol))
-      ),
-      class = "upsetjs_combination"
-    )
+    asCombination(key, c(), type, sets = unlist(strsplit(key, symbol)), cardinality = value, color=cc(key))
   }
   combinations = mapply(
     toCombination,
@@ -241,18 +197,11 @@ fromExpression = function(upsetjs,
 
   sets = list()
   defined_sets = c()
-  for(c in combinations) {
-    for(s in c$setNames) {
+  for (c in combinations) {
+    for (s in c$setNames) {
       if (!(s %in% defined_sets)) {
         defined_sets = c(defined_sets, s)
-        sets[[s]] = structure(list(
-          name = s,
-          type = "set",
-          color = cc(s),
-          elems = c(),
-          cardinality = 0
-        ),
-        class = "upsetjs_set")
+        sets[[s]] = asSet(s, c(), color=cc(s))
       }
       # determine base set based on type and value
       set = sets[[s]]
@@ -282,6 +231,49 @@ fromExpression = function(upsetjs,
   setProperties(upsetjs, props)
 }
 
+#'
+#' extract the sets from a data frame (rows = elems, columns = sets, cell = contained)
+#' @param df the data.frame like structure
+#' @param attributes the optional column list or data frame
+#' @param order.by order intersections by cardinality or degree
+#' @param limit limit the ordered sets to the given limit
+#' @param colors the optional list with set name to color
+#' @param store.elems store the elements in the sets (default TRUE)
+#' @export
+extractSetsFromDataFrame = function(df,
+                         attributes = NULL,
+                         order.by = "cardinality",
+                         limit = NULL,
+                         colors = NULL,
+                         store.elems = TRUE) {
+  stopifnot(is.data.frame(df))
+  stopifnot((
+    is.null(attributes) ||
+      is.data.frame(attributes) ||
+      is.list(attributes) || is.character(attributes)
+  ))
+  stopifnot(order.by == "cardinality" || order.by == "degree")
+  stopifnottype("limit", limit)
+  stopifnot(is.null(colors) || is.list(colors))
+
+  cc = colorLookup(colors)
+
+  elems = rownames(df)
+
+  toSet = function(key) {
+    sub = elems[df[[key]] == TRUE]
+    x = if(store.elems) sub else c()
+     asSet(key, x, cardinality=length(sub), color=cc(key))
+  }
+
+  set_names = setdiff(colnames(df), if (is.character(attributes))
+    attributes
+    else
+      c())
+  sets = lapply(set_names, toSet)
+
+  sortSets(sets, order.by = order.by, limit = limit)
+}
 
 #'
 #' extract the sets from a data frame (rows = elems, columns = sets, cell = contained)
@@ -293,7 +285,10 @@ fromExpression = function(upsetjs,
 #' @param shared a crosstalk shared data frame
 #' @param shared.mode whether on 'hover' or 'click' (default) is synced
 #' @param colors the optional list with set name to color
+#' @param c_type the combination type to use
+#' @param store.elems whether to store the set elements within the structures (set to false for big data frames)
 #' @return the object given as first argument
+#' @importFrom stats aggregate
 #' @examples
 #' df <- as.data.frame(list(a=c(1, 1, 1),b=c(0, 1, 1)),row.names=c('a', 'b', 'c'))
 #' upsetjs() %>% fromDataFrame(df)
@@ -306,7 +301,9 @@ fromDataFrame = function(upsetjs,
                          limit = NULL,
                          shared = NULL,
                          shared.mode = "click",
-                         colors = NULL) {
+                         colors = NULL,
+                         c_type = NULL,
+                         store.elems = TRUE) {
   checkUpSetCommonArgument(upsetjs)
   stopifnot(is.data.frame(df))
   stopifnot((
@@ -318,52 +315,83 @@ fromDataFrame = function(upsetjs,
   stopifnottype('limit', limit)
   stopifnot(shared.mode == "click" || shared.mode == "hover")
   stopifnot(is.null(colors) || is.list(colors))
+  stopifnot(
+    is.null(c_type) ||
+      c_type == "intersection" ||
+      c_type == "union" || c_type == "distinctIntersection" ||
+      c_type == "none"
+  )
 
   cc = colorLookup(colors)
 
+  gen_type = ifelse(!is.null(c_type), c_type, ifelse(isVennDiagram(upsetjs) || isKarnaughMap(upsetjs), 'distinctIntersection', 'intersection'))
+  sorted_sets = extractSetsFromDataFrame(df, attributes, order.by, limit, colors, store.elems = store.elems || gen_type != 'distinctIntersection')
+
   elems = rownames(df)
-  toSet = function(key) {
-    sub = elems[df[[key]] == TRUE]
-    structure(list(
-      name = key,
-      type = "set",
-      color = cc(key),
-      elems = sub
-    ),
-    class = "upsetjs_set")
-  }
-
-  set_names = setdiff(colnames(df), if (is.character(attributes))
-    attributes
-    else
-      c())
-  sets = lapply(set_names, toSet)
-
-  if (!is.null(shared)) {
-    upsetjs = enableCrosstalk(upsetjs, shared, mode = shared.mode)
+  
+  gen = if (!is.null(c_type) && c_type == "none") {
+    list() 
+  } else if (isVennDiagram(upsetjs) || isKarnaughMap(upsetjs)) {
+    if (gen_type == 'distinctIntersection') {
+      extractCombinationsImpl(
+        df,
+        sorted_sets,
+        TRUE,
+        order.by,
+        limit,
+        colors,
+        store.elems = store.elems
+      )
+    } else {
+      generateCombinationsImpl(
+        sorted_sets,
+        gen_type,
+        0,
+        NULL,
+        TRUE,
+        'degree',
+        limit,
+        colors,
+        store.elems = store.elems
+      )
+    }
+  } else if (gen_type == 'distinctIntersection') {
+    extractCombinationsImpl(
+      df,
+      sorted_sets,
+      FALSE,
+      order.by,
+      limit,
+      colors,
+      store.elems = store.elems
+    )
   } else {
-    upsetjs = enableCrosstalk(upsetjs, df, mode = shared.mode)
+    generateCombinationsImpl(
+      sorted_sets,
+      gen_type,
+      0,
+      NULL,
+      FALSE,
+      order.by,
+      limit,
+      colors,
+      store.elems = store.elems
+    )
   }
 
-  sorted_sets = sortSets(sets, order.by = order.by, limit = limit)
-  gen_sets = if (is.null(limit)) sets else sorted_sets
-
-  gen = if (isVennDiagram(upsetjs) || isKarnaughMap(upsetjs)) {
-    generateCombinationsImpl(gen_sets,
-                             'distinctIntersection',
-                             0,
-                             NULL,
-                             TRUE,
-                             'degree',
-                             NULL,
-                             colors)
-  } else {
-    generateCombinationsImpl(gen_sets, 'intersection', 0, NULL, FALSE, order.by, NULL, colors)
+  if (!store.elems && gen_type != 'distinctIntersection') {
+    # delete
+    for(i in 1:length(sorted_sets)) {
+      sorted_sets[[i]]$elems = c()
+    }
   }
-  props = list(sets = sorted_sets,
-               combinations = gen,
-               elems = elems,
-               expressionData = FALSE)
+
+  props = list(
+    sets = sorted_sets,
+    combinations = gen,
+    elems = elems,
+    expressionData = FALSE
+  )
 
   upsetjs = setProperties(upsetjs, props)
 
@@ -375,10 +403,15 @@ fromDataFrame = function(upsetjs,
     upsetjs = setAttributes(upsetjs, attr_df)
   }
 
+  if (!is.null(shared)) {
+    upsetjs = enableCrosstalk(upsetjs, shared, mode = shared.mode)
+  } else {
+    upsetjs = enableCrosstalk(upsetjs, df, mode = shared.mode)
+  }
+
   upsetjs
 }
 
-
 #'
 #' extract the vector of elements
 #' @param upsetjs an object of class \code{upsetjs} or \code{upsetjs_proxy}
@@ -392,6 +425,20 @@ getElements = function(upsetjs) {
   upsetjs$x$elems
 }
 
+#'
+#' set the vector of elements
+#' @param upsetjs an object of class \code{upsetjs}
+#' @param value the vector of elements
+#' @return the object given as first argument
+#' @examples
+#' upsetjs() %>% setElements(c(1,2,3,4,5)) %>% getElements()
+#'
+#' @export
+setElements = function(upsetjs, value) {
+  stopifnot(inherits(upsetjs, 'upsetjs_common'))
+  setProperty(upsetjs, 'elems', value)
+}
+
 #'
 #' extract the vector of sets
 #' @param upsetjs an object of class \code{upsetjs}
@@ -405,6 +452,20 @@ getSets = function(upsetjs) {
   upsetjs$x$sets
 }
 
+#'
+#' set the vector of sets
+#' @param upsetjs an object of class \code{upsetjs}
+#' @param value the vector of sets
+#' @return the object given as first argument
+#' @examples
+#' upsetjs() %>% setCombinations(list(asSet('a', c(1,2,3)))) %>% getSets()
+#'
+#' @export
+setSets = function(upsetjs, value) {
+  stopifnot(inherits(upsetjs, 'upsetjs_common'))
+  setProperty(upsetjs, 'sets', value)
+}
+
 #'
 #' extract the vector of combinations
 #' @param upsetjs an object of class \code{upsetjs}
@@ -418,6 +479,20 @@ getCombinations = function(upsetjs) {
   upsetjs$x$combinations
 }
 
+#'
+#' set the vector of combinations
+#' @param upsetjs an object of class \code{upsetjs}
+#' @param value the vector of combinations
+#' @return the object given as first argument
+#' @examples
+#' upsetjs() %>% setCombinations(list(asCombination('a', c(1,2,3)))) %>% getCombinations()
+#'
+#' @export
+setCombinations = function(upsetjs, value) {
+  stopifnot(inherits(upsetjs, 'upsetjs_common'))
+  setProperty(upsetjs, 'combinations', value)
+}
+
 generateCombinations = function(upsetjs,
                                 c_type,
                                 min,
@@ -435,6 +510,8 @@ generateCombinations = function(upsetjs,
   stopifnot(is.null(limit) ||
               (is.numeric(limit) && length(limit) == 1))
   stopifnot(is.null(colors) || is.list(colors))
+  stopifnot(c_type == "intersection" ||
+              c_type == "union" || c_type == "distinctIntersection")
 
   if (inherits(upsetjs, 'upsetjs_common')) {
     sets = upsetjs$x$sets
@@ -452,6 +529,7 @@ generateCombinations = function(upsetjs,
   }
   setProperty(upsetjs, 'combinations', gen)
 }
+
 #'
 #' configure the generation of the intersections
 #' @param upsetjs an object of class \code{upsetjs} or \code{upsetjs_proxy}
@@ -473,8 +551,16 @@ generateIntersections = function(upsetjs,
                                  order.by = "cardinality",
                                  limit = NULL,
                                  colors = NULL) {
-  generateCombinations(upsetjs, "intersection", min, max, empty, order.by, limit, colors)
+  generateCombinations(upsetjs,
+                       "intersection",
+                       min,
+                       max,
+                       empty,
+                       order.by,
+                       limit,
+                       colors)
 }
+
 #'
 #' configure the generation of the distinct intersections
 #' @param upsetjs an object of class \code{upsetjs} or \code{upsetjs_proxy}
